#!/bin/env perl

use strict;
use lib "Include";
use Config;
use Cwd;
use Sys::Hostname;
use Sys::HostIP;
use IO::Socket;
use XML::Simple;
use Config::IniFiles;
use File::Basename;
use File::Path qw ( mkpath );
use DE_EPAGES::Core::API::Error qw( ExistsError GetError);
use DE_EPAGES::Core::API::File qw( WindowsPath CreateDirectory CopyDirectory WriteFile GetFileContentLines DeleteFile RunOnDirectory );
use DE_EPAGES::Core::API::PerlTools qw( ElementOf GetArchName UsePackage );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Database::API::Connection qw( ConnectSite DisconnectSite GetCurrentDBHandle RunOnStore );
use DE_EPAGES::Database::API::BackupMySQL qw( MyCnfDefault );
use DE_EPAGES::Database::API::MyIniReader;
use DE_EPAGES::Cartridge::API::Table::Cartridge qw( GetAllInfoCartridges );
use DE_EPAGES::XML::API::XML::CartridgeInstallerImportHandler qw( GetCartridges );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::ShopConfiguration::API::StoreConnection qw( GetStoreConnectionParams );

BEGIN {
    if ( $Config{'osname'} eq 'MSWin32' ) {
        UsePackage( "Win32::Process::Info" );
    } else {
        UsePackage( "Proc::ProcessTable" );
    }
}

sub Main {

    local $| = 1;

    my $Cwd = dirname($0);
    chdir $Cwd or die "$Cwd: $!";
    $Cwd = cwd();

    my %PrereqIni;
    tie %PrereqIni, 'Config::IniFiles', (-file => 'prerequisits.ini') or die "prerequisits.ini: $!\n";

    my ($PatchVersion, $BackupDB) =  ReadMakefilePL($Cwd);
    my $HostMk = ReadHostMk($Cwd);

    my ($FirstArg, $Arguments, $SetEnv, $RunScript) = GetArguments(@ARGV);

    SetFilesToChange($HostMk, $PatchVersion, $FirstArg);

    my $StoreConf = ReadDatabaseD();

    my $Errors = "";

    $Errors .= GetDatabaseType($HostMk, $StoreConf, $PrereqIni{EPAGES}{sybase_version}, \$BackupDB);

    $Errors .= TestEpagesVersion($HostMk, $PatchVersion, \%PrereqIni, $StoreConf);

    $Errors .= TestSystemTime($HostMk, 'prerequisits.ini');

    $Errors .= TestArchitecture($HostMk, $PrereqIni{EPAGES}{architecture});

    $Errors .= TestEnvironment($HostMk, \%PrereqIni, $Arguments);

    $Errors .= TestExecutor($HostMk, $PrereqIni{UNIX}{executor}, $RunScript);

    $Errors .= TestApache($HostMk, \%PrereqIni);

    $Errors .= TestDatabase($HostMk, \%PrereqIni, $StoreConf, \$BackupDB);

    $Errors .= TestStores($HostMk, \%PrereqIni, $StoreConf);

    $Errors .= TestLanguagePack($HostMk, $PatchVersion);

    if ($Errors) {
        print STDERR "\nEXITING: Following requirements do not comply:\n";
        print STDERR "==============================================\n";
        $BackupDB = "" unless $BackupDB =~ /[a-z]/s;
        die "$Errors$BackupDB";
    }
    Warning( GetError() ) if ExistsError();

    WriteHostMk($HostMk, $FirstArg);

    if (not -f "makefile" or -M "makefile" > -M "Makefile.PL" ) {
        my $PerlExe = $Config{'osname'} eq 'MSWin32' ? $^X : $ENV{'PERL'};
        my $Command = "$SetEnv $PerlExe -IInclude Makefile.PL 2>&1";
        print "$Command\n";
        system($Command);
        die if $?;
    }

    RunMake(@ARGV);
}

############################################################################
# read makefile
############################################################################

sub ReadMakefilePL {
    my ($Cwd) = @_;

    my ($PatchVersion, $BackupDB);

    foreach my $Line (@{GetFileContentLines("Makefile.PL")}) {
        if ($Line =~ /^\s*PATCH_VERSION\s*[=]\s*(.+)/) {
            $PatchVersion = $1;
        } elsif ($Line =~ /^\s*BACKUP_DB\s*[=]\s*(.+)/) {
            $BackupDB = $1;
        }
    }
    $BackupDB =~ s/\s//s;
    $BackupDB = $BackupDB ? 1 : "";

    die "$Cwd/Makefile.PL: void version setting: PATCH_VERSION=\n" unless $PatchVersion;
    return ($PatchVersion, $BackupDB);
}

sub ReadHostMk {
    my ($Cwd) = @_;

    my %HostMk;
    my $HostName = hostname();
    if (not -f "host-$HostName.mk") {
        $HostMk{PATCH_DIR} = $Cwd;
        $HostMk{PATCH_DIR} = WindowsPath($Cwd) if $Config{'osname'} eq 'MSWin32';
        return \%HostMk;
    }

    open my $HOST_MK, "<", "host-$HostName.mk" or die "$Cwd/host-$HostName.mk: Can't read file: $!";
    while (<$HOST_MK>) {
        /^\s*(.+?)\s*=\s*(.*?)\s*$/ and $HostMk{$1} = $2;
    }
    close $HOST_MK;
    $HostMk{PATCH_DIR} = $Cwd unless $HostMk{PATCH_DIR};
    $HostMk{PATCH_DIR} = WindowsPath($Cwd) if $Config{'osname'} eq 'MSWin32';

    return \%HostMk;
}

sub GetArguments {
    my (@ArgList) = @_;
    return ($ArgList[0], join(' ', @ArgList)) if $Config{osname} eq 'MSWin32';

    my $Arguments = "";
    foreach (@ArgList) {
        if (/[^\w+,.\/-]/) {
            s/\"/\\\"/ if /[\"\']/;
            $Arguments .= /\'/ ? "\"$_\" " : "'$_' ";
        } else {
            $Arguments .= "$_ ";
        }
    }

    my $EpagesProject = $ENV{EPAGES_PROJECT};
    $EpagesProject = 'epages6' unless $EpagesProject;

    my $EtcDir = dirname($ENV{EPAGES}). "/etc";
    $EtcDir = "/etc" unless -d "$EtcDir/init.d";
    my $EpagesInitd = "/etc/init.d/$EpagesProject";
    $EpagesInitd = "$EtcDir/init.d/${EpagesProject}su" if (getpwuid($>))[0] ne 'root';

    my $FirstArg = $ArgList[0];
    $FirstArg =~ s,[/\\],-,g;

    return ($FirstArg, $Arguments, ". /etc/default/$EpagesProject ; ", $EpagesInitd);
}

############################################################################
# set files to change
############################################################################

sub SetFilesToChange {
    my ($HostMk, $PatchVersion, $FirstArg) = @_;

    foreach my $What (qw(WEBADAPTER REQROUTER SERVICE BIN CARTRIDGES CONFIG STORES WEBROOT PERL)) {
        my $Dir = $ENV{"EPAGES_$What"};
        $Dir = "$ENV{EPAGES}/bin" if $What eq "BIN";
        $Dir = "$ENV{EPAGES}/Service" if $What eq "SERVICE";
        $Dir = "$ENV{EPAGES}/RequestRouter" if $What eq "REQROUTER";
        $Dir = "$ENV{EPAGES}/WebAdapter" if $What eq "WEBADAPTER";

        my $Done = $FirstArg =~ /install$/ ? 'patch' : 'restore' ;
        if (-f "$Dir/.$PatchVersion-$Done") {
            $HostMk->{"CHANGE_$What"} = 0;
        } else {
            $HostMk->{"CHANGE_$What"} = -d $Dir ? 1 : 0;
        }
        CheckIntegrityUsed($HostMk, $Dir) if $What eq "CARTRIDGES";
    }
}

sub CheckIntegrityUsed {
    my ($HostMk, $Dir) = @_;

    return if defined $HostMk->{CHANGE_SIC};
    if (not -d $Dir) {
        $HostMk->{CHANGE_SIC} = 0;
        return;
    }

    my $epagesScheduler = "$ENV{EPAGES}/bin/epagesScheduler.sh";
    $epagesScheduler =  $ENV{EPAGES}. '\\bin\\epagesScheduler.cmd' if $Config{osname} eq 'MSWin32';
    my $CronTab = `"$epagesScheduler" show`;

    if ($CronTab !~ /epages/i) {
        $CronTab = `"$epagesScheduler" start`;
        system("\"$epagesScheduler\" stop") and die;
    }
    $HostMk->{CHANGE_SIC} = $CronTab =~ /(Task_|CheckIntegrity)/ ? 1 : 0;
    $HostMk->{STOP_SIC} = '$(PERL) "$(EPAGES)/bin/CheckIntegrity.PL" -del';
    $HostMk->{START_SIC} = '$(PERL) "$(EPAGES)/bin/CheckIntegrity.PL" -add scheduler -debug 9';

    if ($Config{osname} ne 'MSWin32' and $HostMk->{CHANGE_SIC}) {
        if ($CronTab !~ /-g .*CheckIntegrity.+own:epages6/) {
            $HostMk->{START_SIC} = '$(PERL) "$(EPAGES)/bin/CheckIntegrity.PL" -add scheduler=epages -debug 9';
        } elsif ((getpwuid($>))[0] eq 'root' && $CronTab !~ / root .+CheckIntegrity/) {
            if ($CronTab =~ /(\S+\s+){5}(\S+)\s+CheckIntegrity/) {
                $HostMk->{STOP_SIC} =  "\$(SU_EXE) $2 -c '$HostMk->{STOP_SIC}'";
                $HostMk->{STOP_SIC} =  "\$(SU_EXE) $2 -c '$HostMk->{START_SIC}'";
            }
        }
    }

    $HostMk->{CRON_FILES} = '$(EPAGES_CONFIG)/Scheduler.conf ' if $CronTab =~ /epages/;
    if ($CronTab =~ m,.*[ \t]['"]*(.+?/CheckIntegrity.conf),) {
        $HostMk->{CRON_FILES} .= $1;
    }
    return;
}

#========================================================================================
# function     ReadDatabaseD
#----------------------------------------------------------------------------------------
# syntax       $hStoresConf = ReadDatabaseD();
#----------------------------------------------------------------------------------------
# description  Gets a hash of store information from Database.d/*.conf for all stores
#               with section Cartridges and parameters Connection::StoreType,Database
#----------------------------------------------------------------------------------------
# return       $hStoresConf | hash of store information | hash
#========================================================================================
sub ReadDatabaseD () {

    return RunOnDirectory( "$ENV{EPAGES_CONFIG}/Database.d", sub {
            my %StoresConf;

            foreach my $File (glob("*.conf")) {
                my $Store = $File =~ /(.+)\.conf$/ ? $1 : "";
                next unless -r $File;

                my %Ini;
                tie %Ini, 'Config::IniFiles', (-file => $File) or die "$File: $!\n";

                next unless defined $Ini{Cartridges};
                next unless $Ini{Connection}{StoreType};
                next unless $Ini{Connection}{Database};
                next if $Ini{Connection}{StoreType} eq 'Site' and $ENV{EPAGES_IGNORE_SITE};
                $StoresConf{$Store} = { %Ini };
            }

            return \%StoresConf;
        });
}

############################################################################
# get database type
############################################################################

sub GetDatabaseType {
    my ($HostMk, $StoreConf, $SybaseVersion, $BackupDB) = @_;

    return if defined $HostMk->{DB_TYPE_OK};

    my $DataSource = GetDataSource($StoreConf);

    my $Errors = FindSybase($HostMk, $StoreConf, $SybaseVersion, $DataSource);

    $Errors .= FindMySQL($HostMk, $StoreConf, $DataSource, $BackupDB);

    $HostMk->{DB_TYPE_OK} = $Errors ? 0 : 1;
    return $Errors;
}

#---------------------------------------------------------------------------
# find sybase client and/or daemon
#---------------------------------------------------------------------------

sub FindSybase {
    my ($HostMk, $StoreConf, $SybaseVersion, $DataSource) = @_;

    $HostMk->{SYBASE_DATASOURCE} = 1 if grep { $DataSource->{$_} eq 'SYBASE' } keys %$DataSource;

    my $SybaseExe = $ENV{SYBASE}.'\\'.$ENV{SYBASE_ASE}.'\bin\sqlsrvr.exe';
    $SybaseExe = $ENV{SYBASE}.'/'.$ENV{SYBASE_ASE}.'/bin/dataserver' if $Config{osname} ne 'MSWin32';
    $HostMk->{SYBASE_SERVER} = 1 if -f $SybaseExe;
    $HostMk->{CHANGE_SYBASE} = $HostMk->{SYBASE_SERVER};

    $HostMk->{SQL_INI} = $Config{osname} eq 'MSWin32' ? "$ENV{SYBASE}\\ini\\sql.ini" : "$ENV{SYBASE}/interfaces";
    $HostMk->{SYBASE_CLIENT} = 1 if -f $HostMk->{SQL_INI};
    $HostMk->{DB_CLIENT} = 1 if $HostMk->{SYBASE_CLIENT};
    $HostMk->{DB_SERVER} = 1 if $HostMk->{SYBASE_SERVER};

    my $Errors;
    if ($HostMk->{SYBASE_DATASOURCE}) {
        $HostMk->{SYBASE_RUNS} = 1 if SybaseRuns($SybaseExe);
        $Errors = CheckSybaseVersion($SybaseVersion);
    }
    return $Errors;
}

#---------------------------------------------------------------------------
# find running sybase
#---------------------------------------------------------------------------

sub SybaseRuns {
    my ($SybaseExe) = @_;

    return 1 if $ENV{RUNNING_SYBASE_FOUND};
    return 0 unless -f $SybaseExe;

    my $Found = 0;
    if ( $Config{osname} eq 'MSWin32' ) {
        my $pi =  Win32::Process::Info->new();
        foreach my $Proc (@{$pi->GetProcInfo()}) {
            if (lc($Proc->{ExecutablePath}) eq lc($SybaseExe)) { $Found = 1; last; }
        }
    } else {
        my $pi = new Proc::ProcessTable;
        foreach my $Proc (@{$pi->table}) {
            if (0 == index($Proc->cmndline, "$SybaseExe "))
            { $Found = 1; last; }
        }
    }
    return $Found;
}

#---------------------------------------------------------------------------
# check sybase version
#---------------------------------------------------------------------------

sub CheckSybaseVersion {
    my ($RequiredVersion) = @_;
    my $Errors;

    $RequiredVersion =~ s/[^0-9]//g;
    return $Errors unless $RequiredVersion =~ /\S/;

    my $TestFile = "$ENV{SYBASE}/ASEP/bin/ASEPlugin.jpr";
    if (-f $TestFile) {
        open my $CHECK, '<', $TestFile or die "$TestFile: $!";
        my $TestVersion = (split(/=/, (grep {/PluginId=.+/} <$CHECK>)[0]))[1];
        close $CHECK;
        $Errors .= "\nERROR: ASEP version ($TestVersion) must be $RequiredVersion\n"
            if $TestVersion !~ /\Q$RequiredVersion\E/;
    }

    $TestFile = "$ENV{SYBASE}/OCS-15_0/bin/";
    $TestFile .= $Config{osname} eq 'MSWin32' ? 'instjava.exe' : 'installjava';
    if (-f $TestFile) {
        open my $CHECK, '<', $TestFile or die "$TestFile: $!";
        ## no critic (ProhibitNegativeExpressionsInUnlessAndUntilConditions)
        $Errors .= "\nERROR: Wrong version, must be $RequiredVersion: $TestFile\n"
            unless (grep {/\Q$RequiredVersion\E/} <$CHECK>)[0];
        ## use critic
        close $CHECK;
    }
    return $Errors;
}

#---------------------------------------------------------------------------
# find mysql client and/or daemon
#---------------------------------------------------------------------------

sub FindMySQL {
    my ($HostMk, $StoreConf, $DataSource, $BackupDB) = @_;

    $HostMk->{MYSQL_DATASOURCE} = 1 if grep { $DataSource->{$_} eq 'MYSQL' } keys %$DataSource;

    my $Errors;
    foreach my $Store (keys %$StoreConf) {
        next unless $DataSource->{$Store} eq 'MYSQL';
        my $MySQLConfigFile = $StoreConf->{$Store}{Connection}{MySQLConfigFile};
        my $Default = $MySQLConfigFile ? "" : ' default';
        my $Wrong = $MySQLConfigFile ? "Wrong" : 'No';
        $MySQLConfigFile = MyCnfDefault() unless $MySQLConfigFile;
        if (-r $MySQLConfigFile) {
            $HostMk->{MYSQL_CLIENT} = 1 if $Config{osname} eq 'MSWin32';
        } elsif ($$BackupDB) {
            $Errors .= "\nERROR: $Wrong key MySQLConfigFile in Database.d/$Store.conf:\n".
                "  Can't read$Default: $MySQLConfigFile\n";
            $$BackupDB = "\n  Backup/restore of ePages databases is not possible.\n".
                "  You may patch ePages w/o backup by unsetting BACKUP_DB in Makefile.PL\n";
        }
    }

    if ( $Config{osname} eq 'MSWin32' ) {
        $HostMk->{MYSQL_SERVER} = $HostMk->{MYSQL_CLIENT};
    } else {
        $HostMk->{MYSQL_CLIENT} = 1 if system("rpm -q --quiet epages-mysql") == 0;
        $HostMk->{MYSQL_SERVER} = 1 if system("rpm -q --quiet epages-mysqld") == 0;
    }

    $HostMk->{DB_CLIENT} = 1 if $HostMk->{MYSQL_CLIENT};
    $HostMk->{DB_SERVER} = 1 if $HostMk->{MYSQL_SERVER};

    return $Errors;
}

############################################################################
# test epages version
############################################################################

sub TestEpagesVersion {
    my ($HostMk, $PatchVersion, $PrereqIni, $StoreConf) = @_;

    $HostMk->{IS_AS} = ReadWebInterfaceConf($HostMk);

    my ($Errors, @TestCartridges, @ValidVersions, $EpagesVersion);

    $Errors .= CheckPrerequisitsIni($PrereqIni, \@TestCartridges, \@ValidVersions);

    $Errors .= ReadLicenseXml($HostMk, $PrereqIni);

    return $Errors if defined $HostMk->{INITIAL_INTERNAL_VERSION};

    if ($HostMk->{IS_AS}) {
        $Errors .= FindVoidCartridgeSections($StoreConf);
        $Errors .= FindVoidCartridges($StoreConf, $PrereqIni->{EPAGES}{void_cartridges});
        $Errors .= GetVersionFromStoreConf($PrereqIni, $StoreConf, \@TestCartridges, \@ValidVersions, \$EpagesVersion);

        $EpagesVersion = $ValidVersions[0] unless $EpagesVersion;
        push @ValidVersions, $PatchVersion;
        SetEpagesVersion($HostMk, $EpagesVersion, \@ValidVersions);
        return $Errors;
    }

    return;
}

#---------------------------------------------------------------------------
# read WebInterface.conf
#---------------------------------------------------------------------------

sub ReadWebInterfaceConf {
    my ($HostMk) = @_;

    return $HostMk->{IS_AS} if defined $HostMk->{IS_AS};

    my $ePagesService = "$ENV{EPAGES}/Service/epagesService";
    $ePagesService .= ".exe" if $Config{osname} eq 'MSWin32';
    my $WIConf = "$ENV{EPAGES_CONFIG}/WebInterface.conf";

    return 0 if !$HostMk->{DB_CLIENT} || !-e $ePagesService || !-e $WIConf;

    my %Ini;
    tie %Ini, 'Config::IniFiles', (-file => $WIConf, -nocase => 1) or die "$WIConf: $!\n";

    my $IPs = Sys::HostIP->ips;
    push @$IPs, inet_ntoa( scalar( gethostbyname( hostname())));

    foreach my $SectionName (keys %Ini) {
        next unless @$IPs ~~ $SectionName;
        next unless $Ini{$SectionName}{applicationserver};
        return 1;
    }

    return 0;
}

#---------------------------------------------------------------------------
# check prerequisits.ini
#---------------------------------------------------------------------------

sub CheckPrerequisitsIni {
    my ($PrereqIni, $TestCartridges, $ValidVersions) = @_;
    my ($Errors);

    $Errors .= 'test_cartridges ' unless $PrereqIni->{EPAGES}{test_cartridges};
    @$TestCartridges = split( /,/, $PrereqIni->{EPAGES}{test_cartridges});

    $Errors .= 'epages_version ' unless $PrereqIni->{EPAGES}{epages_version};
    @$ValidVersions = split( /,/, $PrereqIni->{EPAGES}{epages_version});

    $Errors = "\nERROR: prerequisits.ini:\n  Following keys in section 'EPAGES' missing:\n$Errors\n"
        if $Errors;

    my $MissingSections = "";
    map { $MissingSections .= "$_ " unless $PrereqIni->{$_}; } @$TestCartridges;
    chop($MissingSections);

    $Errors .= "\nERROR: prerequisits.ini:\n  Following sections missing:\n$MissingSections\n"
        if $MissingSections;
    return $Errors;
}

#---------------------------------------------------------------------------
# read License.xml
#---------------------------------------------------------------------------

sub ReadLicenseXml {
    my ($HostMk, $PrereqIni) = @_;

    return unless $HostMk->{IS_AS};
    return if $HostMk->{HAS_EPAGES_LICENSE};
    return unless $PrereqIni->{EPAGES}{epages_license};

    my $ref = XMLin("$ENV{EPAGES_CONFIG}/License.xml") or
        return "\nERROR: Cannot read file:\nXMLin($ENV{EPAGES_CONFIG}/License.xml)\n";

    my $MajorRelease;
    foreach my $License (@{$ref->{Content}{License}}) {
        my @ConditionList = ref $License->{Condition} ?
            @{$License->{Condition}} : ($License->{Condition});
        foreach my $Condition (@ConditionList) {
            if ($Condition->{Name} eq 'MajorRelease') {
                $MajorRelease = $Condition->{Value};
                $HostMk->{HAS_EPAGES_LICENSE} = 1 if $MajorRelease eq $PrereqIni->{EPAGES}{epages_license};
            }
        }
    }
    print "TestEpagesVersion: License.xml: MajorRelease: $MajorRelease\n";
    return $HostMk->{HAS_EPAGES_LICENSE} ? "" :
        "\nERROR: License.xml: MajorRelease must be '$PrereqIni->{EPAGES}{epages_license}', not '$MajorRelease'\n";
}

#---------------------------------------------------------------------------
# find stores without parameter DE_EPAGES::Trigger in section Cartridges
#---------------------------------------------------------------------------

sub FindVoidCartridgeSections {
    my ($StoreConf) = @_;

    my (@aVoidSections);

    foreach my $Store (keys %$StoreConf) {
        print "TestEpagesVersion: Database.d/$Store.conf, key 'DE_EPAGES::Trigger'\n";
        push(@aVoidSections, "$Store.conf") unless $StoreConf->{$Store}{Cartridges}{'DE_EPAGES::Trigger'};
    }

    my $VoidSections = join(', ', map { "'$_'" } @aVoidSections );
    return $VoidSections ? "\nERROR: Files w/o valid DE_EPAGES::Trigger key-value pair:\n$VoidSections\n" : "";
}

#---------------------------------------------------------------------------
# find void cartridges
#---------------------------------------------------------------------------

sub  FindVoidCartridges {
    my ($StoreConf, $VoidItem) = @_;

    return "" unless $VoidItem;
    return "" unless scalar keys %$StoreConf;
    my @VoidCartridges = split( /,/, $VoidItem);

    my $Errors;
    foreach my $Store (keys %$StoreConf) {
        print "TestEpagesVersion: $Store: ", join(',', @VoidCartridges). "\n";
         foreach (@VoidCartridges) {
             $Errors .= "DB $Store: cartridge '$_' mustn't be installed\n" if defined $StoreConf->{$Store}{Cartridges}{$_};
         }
    }
    return $Errors ? "\nERROR: Cartridge(s) mustn't be installed in any database:\n$Errors\n" : "";
}

#---------------------------------------------------------------------------
# check cartridge versions
#---------------------------------------------------------------------------

sub GetVersionFromStoreConf {
    my ($PrereqIni, $StoreConf, $TestCartridges, $ValidVersions, $EpagesVersion) = @_;

    return "" unless scalar keys %$StoreConf;

    my $Errors = "";
    foreach my $Store (keys %$StoreConf) {
        foreach my $Cartridge (@$TestCartridges) {
            my $CartridgeVersion = $StoreConf->{$Store}{Cartridges}{$Cartridge} or next;
            my $Version = $PrereqIni->{$Cartridge}{$CartridgeVersion};
            print "TestEpagesVersion: $Store: $Cartridge: $CartridgeVersion (ePages $Version)\n";
            if (not $Version or not @$ValidVersions ~~ $Version) {
                $Errors .= "DB $Store, cartridge $Cartridge, wrong version: $CartridgeVersion (epages: $Version)\n";
            } elsif (not $$EpagesVersion) {
                $$EpagesVersion = $Version;
            }
        }
    }
    return $Errors ? "\nERROR: Version mismatch in some databases:\n$Errors\n" : "";
}

#---------------------------------------------------------------------------
# set epages version
#---------------------------------------------------------------------------

sub SetEpagesVersion {
    my ($HostMk, $EpagesVersion, $ValidVersions) = @_;
    my $Initial;

    ($HostMk->{INITIAL_MAIN_VERSION},  $HostMk->{INITIAL_SUB_VERSION},
     $HostMk->{INITIAL_PATCH_LEVEL},   $HostMk->{INITIAL_SERIAL_NUMBER},
     $HostMk->{INITIAL_MAJOR_VERSION}, $Initial,
     $HostMk->{INITIAL_INTERNAL_VERSION}) = SplitEpagesVersion($EpagesVersion);

    $HostMk->{INITIAL_VERSION} = $EpagesVersion;
    $HostMk->{$Initial} = 1;

    foreach (@$ValidVersions) {
        my $ValidInternal = (SplitEpagesVersion($_))[6];
        $HostMk->{"INITIAL_VERSION_LT_$ValidInternal"} = 1 if $HostMk->{INITIAL_INTERNAL_VERSION} < $ValidInternal;
    }
}

sub SplitEpagesVersion {
    my ($EpagesVersion) = @_;

    my ($Main, $Sub, $Level, $Serial) = split(/\./, $EpagesVersion);
    $_ ||= 0  foreach ($Main, $Sub, $Level, $Serial);
    my $Major = "$Main.$Sub";
    s/[^0-9.]//g foreach ($Major, $Main, $Sub, $Level, $Serial);
    my $Initial = "INITIAL_EPAGES". $Main;
    if ( $EpagesVersion eq "6.0beta" ) { $Main = 5; $Sub = 9; }

    my $Internal = sprintf "%d%.3d%.3d%.3d", $Main, $Sub, $Level, $Serial;

    return ($Main, $Sub, $Level, $Serial, $Major, $Initial, $Internal);
}

############################################################################
# test system time
############################################################################

sub TestSystemTime {
    my ($HostMk, $IniFile) = @_;

    return if defined $HostMk->{SYSTEM_TIME_OK};

    my $t = time;
    open my $FILE, ">", ".empty" or die ".empty: $!";
    close($FILE);

    my $TimeString = scalar localtime($t);
    print "TestSystemTime: scalar localtime: $TimeString\n";

    my $Errors = "";
    if (-M ".empty" > -M $IniFile) {
        $Errors = "\nERROR: $IniFile: File is newer than system time ($TimeString).\n".
            "Probably your system time is inaccurate, adjust it.\n".
            "If you do not want to adjust your system time, touch the file.\n";
    }
    $HostMk->{SYSTEM_TIME_OK} = $Errors ? 0 : 1;
    return $Errors;
}

############################################################################
# test architecture
############################################################################

sub TestArchitecture {
    my ($HostMk, $TestItem) = @_;

    return if defined $HostMk->{ARCHITECTURE_OK};
    return unless $TestItem =~ /\S/;

    my ($ArchName, $ArchName32) = GetArchName();
    print "TestArchitecture: GetArchName(): $ArchName\n";

    my @SupportedArchitectures = split( /,/, $TestItem );

    my $Errors = "";
    $Errors = "\nERROR: $ArchName: perl architecture not supported, patch requires one of:\n$TestItem\n"
        unless @SupportedArchitectures ~~ $ArchName;
    $HostMk->{ARCHITECTURE_OK} = $Errors ? 0 : 1;
    return $Errors;
}

############################################################################
# test environment
############################################################################

sub TestEnvironment {
    my ($HostMk, $PrereqIni, $Arguments) = @_;

    return if defined $HostMk->{ENVIRONMENT_OK};

    my $Errors = "";

    if ($Config{osname} ne 'MSWin32') {
        my $FileType = `file "$ENV{EPAGES_PERL}/bin/perl"`;
        print "TestEnvironment: $FileType";
        my $Run = -f "patch.sh" ? ", better run:\n  ./patch.sh $Arguments\n" : "\n";
        $Errors .= "\nERROR: $ENV{EPAGES_PERL}/bin/perl is not a shell script$Run"
            if $FileType =~ /ELF 32-bit/;
    }
    my $TestItem = $PrereqIni->{EPAGES}{perl_version};

    if ($TestItem =~ /\S/) {
        print "TestEnvironment: Perl version: $Config{version}\n";
        $Errors .= "\nERROR: perl version must be '$TestItem', not '$Config{version}':\n$^X\n"
            if $Config{version} ne $TestItem;
    }

    $TestItem = $PrereqIni->{EPAGES}{environment};
    print "TestEnvironment: $TestItem\n";

    my $UnknownVariables = "";
    map { $UnknownVariables .= "$_ " if $_ and not defined $ENV{$_}; } split( /,/, $TestItem );
    chop($UnknownVariables);

    $Errors .= "\nERROR: Environment variables required for patch are not defined:\n".
        "$UnknownVariables\n" if $UnknownVariables;

    my $EpagesLog = "". $ENV{EPAGES_LOG};
    if ( $EpagesLog =~ /\S/ and not -e $EpagesLog ) {
        print "TestEnvironment: CreateDirectory($EpagesLog)\n";
        CreateDirectory($EpagesLog) or
            $Errors .= "\nERROR: $EpagesLog: Can't create directory\n";
    }
    $HostMk->{ENVIRONMENT_OK} = $Errors ? 0 : 1;
    return $Errors;
}

############################################################################
# test executor
############################################################################

sub TestExecutor {
    my ($HostMk, $TestItem, $RunScript) = @_;

    return if $Config{osname} eq 'MSWin32';
    return if defined $HostMk->{EXECUTOR_OK};
    $HostMk->{EXECUTOR} = (getpwuid($>))[0];

    print "TestExecutor: (getpwuid($>))[0]: $HostMk->{EXECUTOR}\n";

    return unless $TestItem =~ /\S/;

    my $UnsupportedUser = 1;
    map { $UnsupportedUser = 0 if $_ eq $HostMk->{EXECUTOR} } split( /,/, $TestItem );

    my $Errors = "";
    $Errors = "\nERROR: $HostMk->{EXECUTOR}: user not supported, patch must be run by one of: $TestItem\n"
        if $UnsupportedUser;

    my $EpagesProject = $ENV{EPAGES_PROJECT};
    $EpagesProject = 'epages6' unless $EpagesProject;

    if ($HostMk->{EXECUTOR} ne 'root') {
        $Errors .= "\nERROR: $HostMk->{EXECUTOR}: variable EPAGES_ACL=1 required in /etc/default/$EpagesProject\n"
            unless $ENV{EPAGES_ACL};
    }

    if ($ENV{EPAGES_ACL}) {
        print "TestExecutor: setfacl -m u:ep_appl:rwx $ENV{EPAGES}/acltest\n";
        WriteFile("$ENV{EPAGES}/acltest", \1, { binary=>1 } );
        system ("setfacl -m u:ep_appl:rwx $ENV{EPAGES}/acltest");
        $Errors .= "\nERROR: setfacl: variable EPAGES_ACL is set, but command failed ($?)\n"
            if $? and $HostMk->{EXECUTOR} ne 'root';
        DeleteFile("$ENV{EPAGES}/acltest");
    }

    if ($RunScript =~ /\S/) {
        print "TestExecutor: -x $RunScript\n";
        $Errors .= "\nERROR: $HostMk->{EXECUTOR}: can't execute: $RunScript\n"  unless -x $RunScript;
    }

    $HostMk->{EXECUTOR_OK} = $Errors ? 0 : 1;
    return $Errors;
}

############################################################################
# test apache
############################################################################

sub TestApache {
    my ($HostMk, $PrereqIni) = @_;
    my $Errors;

    return $Errors if $Config{osname} eq 'MSWin32';
    return $Errors if defined $HostMk->{APACHE_OK};

    my $EpagesServer = "$ENV{HTTPD_ROOT}/bin/httpd";

    if (-x $EpagesServer) {
        my $Kernel = (split / /, $Config{myuname})[2];
        $Kernel =~ s/^([0-9]+\.[0-9]+\.[0-9]+)\..*/${1}s/;
        $Kernel =~ s/^([0-9]+\.[0-9]+\.[0-9]+)-.*/${1}r/;

        my $EEP = $ENV{EPAGES_PERL};
        my $LDLP = "$ENV{HTTPD_ROOT}/lib:$EEP/lib32/$Kernel:$EEP/lib32/ssl:$EEP/bin:$ENV{LD_LIBRARY_PATH}";

        my $EpagesApaVer = `LD_LIBRARY_PATH="$LDLP" $EpagesServer -v`;
        if ($EpagesApaVer =~ /\/(\S+)/) {
           $EpagesApaVer = $1;
        }
        print "TestApache: $EpagesApaVer\n";

        my $TestItem = $PrereqIni->{EPAGES}{apache_version};
        if ($TestItem =~ /\S/) {
            my $ValidVersionFound = 0;
            foreach (split( /,/, $TestItem)) {
                $ValidVersionFound = 1 if $EpagesApaVer =~ /^\Q$_\E/;
            }
            $Errors = "\nERROR: $EpagesServer: $EpagesApaVer\nApache version must one of: $TestItem\n"
                unless $ValidVersionFound;
        }
    }

    $HostMk->{APACHE_OK} = $Errors ? 0 : 1;
    return $Errors;
}

############################################################################
# test database
############################################################################

sub TestDatabase {
    my ($HostMk, $PrereqIni, $StoreConf, $BackupDB) = @_;

    return "" if defined $HostMk->{DATABASE_OK};

    my $DataSource = GetDataSource($StoreConf);

    my $Errors;

    my %MyCnfFiles;
    $Errors .= CheckStoreConnection($_, $StoreConf, $PrereqIni, $DataSource->{$_}, $HostMk, \%MyCnfFiles, $BackupDB) foreach keys %$StoreConf;

    $HostMk->{DATABASE_OK} = $Errors ? 0 : 1;
    return $Errors;
}

#---------------------------------------------------------------------------
# check store connection
#---------------------------------------------------------------------------

sub CheckStoreConnection {
    my ($Store, $StoreConf, $PrereqIni, $DS, $HostMk, $MyCnfFiles, $BackupDB) = @_;

    my $SectionHost = GetSectionHost($Store, $StoreConf);
    return unless $SectionHost;

    my $Errors;
    my $UnsetBackupDB = "\n  Backup/restore of ePages databases is not possible.\n".
        "  You may patch ePages w/o backup by unsetting BACKUP_DB in Makefile.PL\n";

    if ($$BackupDB) {
        $Errors = CreateDumpDirectory($Store, $StoreConf);
        $$BackupDB = $UnsetBackupDB if $Errors;
    }

    # check BackupType
    my ($LogBackup, $FullBackup);
    foreach (qw(InitialBackupType FinalBackupType BackupType)) {
        $Errors .= "\nERROR: $SectionHost: $StoreConf->{$Store}{Connection}{$_}: void $_, use -full or -log\n"
            if $StoreConf->{$Store}{Connection}{$_} =~ /\S/ and $StoreConf->{$Store}{Connection}{$_} !~ /^-(log|full)$/;
        $LogBackup = 1 if $StoreConf->{$Store}{Connection}{$_} eq '-log';
        $FullBackup = 1 if $StoreConf->{$Store}{Connection}{$_} eq '-full';
    }

    # go on if host is localhost and no database runs on myself
    next if $SectionHost eq 'localhost' and not $HostMk->{"${DS}_SERVER"};

    if ($SectionHost eq hostname() and not $HostMk->{"${DS}_CLIENT"} and $$BackupDB) {
        $Errors .= "\nERROR: $ENV{EPAGES_CONFIG}/Database.d/$Store.conf:\n".
            "Parameter BackupHost found, but no \L$DS\E client on $SectionHost\n";
        $$BackupDB = $UnsetBackupDB;
    }

    # check MySQL backup variant

    if ($DS eq 'MYSQL' and $$BackupDB) {
        my $BackupVariant = $StoreConf->{$Store}{Connection}{BackupVariant};
        my $MySQLConfigFile = $StoreConf->{$Store}{Connection}{MySQLConfigFile};
        my $BackupMySQLArgs = $StoreConf->{$Store}{Connection}{BackupMySQLArgs};

        my $VoidConf;
        unless (defined $MyCnfFiles->{$MySQLConfigFile}) {
            $MyCnfFiles->{$MySQLConfigFile} = 1;
            $VoidConf = CheckMySQLConfiguration('', $MySQLConfigFile, $Store);
            unless ($VoidConf) {
                $VoidConf .= CheckMySQLConfiguration('log_bin', $MySQLConfigFile, $Store) if $LogBackup;
                $VoidConf .= CheckMySQLConfiguration('innodb_data_home_dir', $MySQLConfigFile, $Store) if $FullBackup && $BackupVariant eq 'Hot';
            }
        }
        $VoidConf .= CheckLVM($BackupMySQLArgs) if $FullBackup and $BackupVariant eq 'LVM';
        $$BackupDB = $UnsetBackupDB if $VoidConf;
        $Errors .= $VoidConf;
    }

    my $SpHelpDB = $DS eq 'SYBASE' ? 'sp_helpdb' : 'show databases';
    print "TestDatabase ($DS): ConnectSite($Store); execute('$SpHelpDB')\n";
    eval { ConnectSite($Store); };
    if (ExistsError()) {
        $Errors .= "\nERROR: $SectionHost: Database connection failed: $Store\n".
            "    Check your configuration files:\n".
            "    $ENV{EPAGES_CONFIG}/Database.d/$Store.conf\n".
            "    $HostMk->{SQL_INI}\n\n";
        $Errors .= GetError->message. "\n";
        DisconnectSite();
        next;
    }
    my $dbh = GetCurrentDBHandle();

    $Errors .= CheckDatabaseSpace($PrereqIni->{sysusages}{freespace}, $SectionHost, $Store) if $DS eq 'SYBASE';
    $Errors .= CheckDatabaseOptions($dbh, $PrereqIni, $SectionHost, $Store, $StoreConf->{$Store}{Connection}{Database}, $DS);
    DisconnectSite();

    return $Errors;
}

#---------------------------------------------------------------------------
# check MySQL configuration file
#---------------------------------------------------------------------------

sub CheckMySQLConfiguration {
    my ($Param, $MySQLConfigFile, $Store) = @_;

    $MySQLConfigFile = MyCnfDefault() unless $MySQLConfigFile;
    print "TestDatabase (MYSQL): $MySQLConfigFile\n" unless $Param;

    my $MyCnf;
    eval { $MyCnf = DE_EPAGES::Database::API::MyIniReader->new(FileName => $MySQLConfigFile); };
    if (ExistsError()) {
        my $ErrorMsg = GetError->message;
        $ErrorMsg =~ s/\s+/ /gs;
        $ErrorMsg =~ s/\. at .* line [0-9]+\.//g;
        $ErrorMsg =~ s/ (Line [0-9]+ in file )/\n  $1/g;
        return "\nERROR: $ErrorMsg";
    }
    return "" unless $Param;

    my $Directory = $MyCnf->get( 'mysqld', $Param );
    $Directory =~ s/^'(.*?)'$/$1/;
    $Directory =~ s/^"(.*?)"$/$1/;

    my $Msg = $Param eq 'log_bin' ? "Type is -log" : "Variant is Hot";
    return "\nERROR: Backup$Msg, but MySQL config has no parameter $Param in section [mysqld]\n".
        "Store config: $ENV{EPAGES_CONFIG}/Database.d/$Store.conf\n".
        "MySQL config: $MySQLConfigFile\n" unless $Directory;
    return -d $Directory ? "" :
        "\nERROR: Backup$Msg, but directory does not exist: section [mysqld], key $Param\n".
        "$Param: $Directory:\n".
        "Store config: $ENV{EPAGES_CONFIG}/Database.d/$Store.conf\n".
        "MySQL config: $MySQLConfigFile\n";
}

#---------------------------------------------------------------------------
# check LVM
#---------------------------------------------------------------------------

sub CheckLVM {
    my ($BackupMySQLArgs) = @_;

    return "\nERROR: BackupVariant LVM not supported under Windows.\n" if $Config{'osname'} eq 'MSWin32';
    return "\nERROR: BackupVariant is LVM, but no parameter BackupMySQLArgs defined.\n" unless $BackupMySQLArgs;

    my $VolGroup = $BackupMySQLArgs;
    $VolGroup =~ s/.*-lvmgroup\s+//;
    $VolGroup =~ s/'(.*?)'.*/$1/ or $VolGroup =~ s/"(.*?)".*/$1/ or $VolGroup =~ s/\s.*//;
    return "\nERROR: BackupVariant is LVM, but -lvmgroup in parameter BackupMySQLArgs missing.\n" unless $VolGroup;

    my $LogVol = $BackupMySQLArgs;
    $LogVol =~ s/.*-lvmdata\s+//;
    $LogVol =~ s/'(.*?)'.*/$1/ or $LogVol =~ s/"(.*?)".*/$1/ or $LogVol =~ s/\s.*//;
    return "\nERROR: BackupVariant is LVM, but -lvmdata in parameter BackupMySQLArgs missing.\n" unless $LogVol;

    my $Errors;
    foreach (qw(create remove)) {
        print "TestDatabase (MYSQL): lv$_\n";
        $Errors .= "\nERROR: BackupVariant is LVM, but no executable 'lv$_' found in PATH\n".
            "PATH: $ENV{PATH}\n" unless ExecutableInPath("lv$_");
    }
    return $Errors if $Errors;

    print "TestDatabase (MYSQL): vgdisplay -As | grep -q '^ *\"$VolGroup\"'\n";
    if (system("PATH=\$PATH:/usr/sbin vgdisplay -As | grep -q '^ *\"$VolGroup\"'") == 0) {
        print "TestDatabase (MYSQL): lvs --noheadings -o lv_name $VolGroup | grep -q '^ *$LogVol\$'\n";
        return system("PATH=\$PATH:/usr/sbin lvs --noheadings -o lv_name $VolGroup | grep -q '^ *$LogVol\$'") == 0
            ? "" : "\nERROR: $LogVol: no such logical volume found in group $VolGroup\n";
    } else {
        return "\nERROR: $VolGroup: no such active volume group found\n";
    }
}

#---------------------------------------------------------------------------
# find executable file in path
#---------------------------------------------------------------------------

sub ExecutableInPath {
    my ($File) = @_;
    foreach my $Dir (split(/:/, $ENV{PATH})) {
        return 1 if -x "$Dir/$File";
    }
    return 0;
}

#---------------------------------------------------------------------------
# get datasource from Store.conf
#---------------------------------------------------------------------------

sub GetDataSource {
    my ($StoreConf) = @_;

    my $MyHost = hostname();
    my $DataSource = {};
    foreach my $Store (keys %$StoreConf) {
        my $SectionHost = GetSectionHost($Store, $StoreConf);
        next unless defined $SectionHost;
        $DataSource->{$Store} = $StoreConf->{$Store}{Connection}{datasource};
        $DataSource->{$Store} = $DataSource->{$Store} =~ /dbi:(\w+)/ ? uc($1) : "";
    }
    return $DataSource;
}

#---------------------------------------------------------------------------
# create dump directory
#---------------------------------------------------------------------------

sub CreateDumpDirectory {
    my ($Store, $StoreConf) = @_;

    my $Errors;
    my $DumpDir = $StoreConf->{$Store}{Connection}{DumpDir};
    $DumpDir =~ s/^_(.+?)_/$ENV{$1}/g;
    $DumpDir =~ s/\$ENV\{(.+?)\}/$ENV{$1}/g;

    # create dump dir if necessary
    if ( $DumpDir =~ /\S/ and not -e $DumpDir ) {
        print "TestDatabase: CreateDirectory($DumpDir)\n";
        CreateDirectory($DumpDir) or
            $Errors .= "\nERROR: $DumpDir: Can't create directory\n";
    }
    return $Errors;
}

#---------------------------------------------------------------------------
# get hostname of section in Store.conf
#---------------------------------------------------------------------------

sub GetSectionHost {
    my ($Store, $StoreConf) = @_;

    # check all sections starting with 'host '
    my $SectionHost = $StoreConf->{$Store}{Connection}{BackupHost};
    return "" unless $SectionHost;

    # go on if host isn't localhost or myself
    return unless $SectionHost eq hostname() or $SectionHost =~ /^(ipv6-)?localhost$/;

    return $SectionHost;
}

#---------------------------------------------------------------------------
# check database space
#---------------------------------------------------------------------------

sub CheckDatabaseSpace {
    my ($FreeSpace, $SectionHost, $Connection) = @_;

    # check for each database running on my server if there is enough space in it
    my $Threshold = "";
    $Threshold = 100 - $FreeSpace if $FreeSpace =~ /\S/;

    my $PerlExe = $Config{'osname'} eq 'MSWin32' ? $^X : $ENV{'PERL'};
    my $Command = "\"$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Database/Scripts/SybaseSpaceUsed.pl\"".
        " -storename $Connection -threshold \"$Threshold\"";
    print "TestDatabase (SYBASE): \"$PerlExe\" -IInclude $Command\n";
    my $Errors = `"$PerlExe" -IInclude $Command -nooutput 2>&1`;
    return $Errors ? "\nERROR: $SectionHost: $Connection: Following databases have less than $FreeSpace% free space:\n$Errors" : "";
}

#---------------------------------------------------------------------------
# check database options
#---------------------------------------------------------------------------

sub CheckDatabaseOptions {
    my ($dbh, $PrereqIni, $SectionHost, $Store, $DBName, $DS) = @_;
    my $Errors;

    my $RequiredVersion = $PrereqIni->{EPAGES}{sybase_version};

    if ($RequiredVersion =~ /\S/ and $DS eq 'SYBASE') {
        my $Result = $dbh->execute('select @@version', []);
        my $TestVersion = (split '/', $Result->[0]->[0])[1];
        $Errors .= "\nERROR: $SectionHost: $Store: ASE version ($TestVersion) must be $RequiredVersion\n"
            if $TestVersion ne $RequiredVersion;
    }

    # execute sp_helpdb on my server
    my $SpHelpDB = $DS eq 'SYBASE' ? 'sp_helpdb' : 'show databases';
    my $Result = $dbh->execute( $SpHelpDB, [] , ["String", "String", "String", "Integer", "String", "String"] );

    foreach (@$Result) {
        my ($name, $db_size, $owner, $dbid, $created, $status) = @$_;
        $status = ", $status, ";
        $name =~ s/\s+$//;
        $owner =~ s/\s+$//;
        next unless $DBName eq $name;

        if ( $DS eq 'SYBASE') {
            next unless $owner eq 'sa';
            foreach my $Option (keys %{$PrereqIni->{sp_dboption}}) {
                next unless $PrereqIni->{sp_dboption}{$Option} =~ /^any|$name$/;
                # check for each database mentioned in section if is has req. options
                $Errors .= "\nERROR: $SectionHost: $Store: $Option: option not found in DB $name\n"
                    unless $status =~ /,+\s+$Option,+\s+/;
            }
        } else {
            print "TestDatabase ($DS): ConnectSite($Store); execute('SHOW CREATE DATABASE $name')\n";
            my $Result2 = $dbh->execute("SHOW CREATE DATABASE $name", [] , ["String", "String"] );
            my ($name2, $db_charset) = @{$Result2->[0]};
            $db_charset =~ s/.* (DEFAULT CHARACTER SET.+?) \*.*/$1/;
            # check that DEFAULT CHARACTER SET is utf8
            $Errors .= "\nERROR: $Store: $name2: $db_charset\nExecute: ALTER DATABASE $name CHARACTER SET utf8 COLLATE utf8_bin;\n"
                unless $db_charset =~ /DEFAULT CHARACTER SET utf8/;
        }
    }
    return $Errors;
}

############################################################################
# test stores
############################################################################

sub TestStores {
    my ($HostMk, $PrereqIni, $StoreConf) = @_;
    my $Errors = "";

    return $Errors if defined $HostMk->{STORES_OK};
    return $Errors unless $HostMk->{IS_AS};

    my %StoresInDatabase;

    $Errors .= CheckSiteConnections($HostMk, $StoreConf, \%StoresInDatabase);

    $Errors .= CheckStoreConnections($HostMk, $PrereqIni, $StoreConf, \%StoresInDatabase);

    delete $HostMk->{LANGUAGES_OK} unless $HostMk->{LANGUAGES};
    $HostMk->{LANGUAGES} = join(' ', @{GetUsedLanguages([keys %StoresInDatabase])})
        unless $HostMk->{LANGUAGES_OK};

    $HostMk->{STORES_OK} = $Errors ? 0 : 1;
    return $Errors;
}

#---------------------------------------------------------------------------
# test site connection
#---------------------------------------------------------------------------

sub CheckSiteConnections {
    my ($HostMk, $StoreConf, $StoresInDatabase) = @_;
    my $Errors = "";

    return $Errors if $ENV{EPAGES_IGNORE_SITE};

    my @StoresInDatabaseIni = (keys %$StoreConf);

    print "TestStores: ConnectSite(Site); \$System->get('Stores')\n";
    eval { ConnectSite('Site'); };
    if (ExistsError()) {
        $Errors .= "\nERROR: Database connection failed: Site\n".
            "Start the database server and/or check your configuration files:\n".
            "    $ENV{EPAGES_CONFIG}/Database.d/Site.conf\n".
            "    $HostMk->{SQL_INI}\n\n".
            GetError->message. "\n";
    } else {
        my $System = LoadRootObject();
        foreach my $Store ( @{$System->get('Stores')} ) {
            my $StoreName = $Store->{Attributes}->{Name};
            my $hConnection = GetStoreConnectionParams($StoreName);
            $StoresInDatabase->{$StoreName} = $hConnection->{'active'};
            next unless $hConnection->{'active'};
            $Errors .= "\nERROR: $ENV{EPAGES_CONFIG}/Database.d/$StoreName.conf\nParameter StoreType missing\n"
                unless @StoresInDatabaseIni ~~ $StoreName;
        }
    }
    DisconnectSite();
    return $Errors;
}

#---------------------------------------------------------------------------
# test store connection
#---------------------------------------------------------------------------

sub CheckStoreConnections {
    my ($HostMk, $PrereqIni, $StoreConf, $StoresInDatabase) = @_;
    my $Errors = "";

    my $CartridgesInXmlFile = GetCartridges("$ENV{EPAGES_CARTRIDGES}/Cartridges.xml");

    my %AllStoreTypes;
    foreach my $Store (keys %$StoreConf) {
        print "TestStores: ConnectSite($Store); GetAllInfoCartridges()\n";

        my $IsActive = $StoresInDatabase->{$Store};
        $IsActive = $StoreConf->{$Store}{Connection}{active} if $ENV{EPAGES_IGNORE_SITE};

        unless ($IsActive) {
            my $HasError = defined $IsActive ? 'is not active' : 'does not exist';
            $Errors .= "\nERROR: Store '$Store': $HasError\n";
            next;
        }

        eval { ConnectSite($Store); };
        if (ExistsError()) {
            $Errors .= "\nERROR: Store '$Store': Database connection failed\n".
                "Start the database server and/or check your configuration files:\n".
                "    $ENV{EPAGES_CONFIG}/Database.d/$Store.conf\n".
                "    $HostMk->{SQL_INI}\n\n".
                GetError->message. "\n" if $Store ne 'Site';
            DisconnectSite();
            next;
        }
        my $CartridgesInDatabase = GetAllInfoCartridges();
        DisconnectSite();

        my $StoreType = $StoreConf->{$Store}{Connection}{StoreType};
        next if defined $AllStoreTypes{$StoreType};
        $AllStoreTypes{$StoreType} = 1;

        foreach my $Cartridge (@$CartridgesInDatabase) {
            next if ElementOf($Cartridge->{Package}, split(/,/, $PrereqIni->{EPAGES}{notest_cartridges}));
            $Errors .= "\nERROR: $ENV{EPAGES_CARTRIDGES}/Cartridges.xml: DatabaseType \"$StoreType\": $Cartridge->{Package}: entry missing\n"
                unless $Cartridge->{Package} ~~ @{$CartridgesInXmlFile->{$StoreType}};
        }
    }
    return $Errors;
}

#========================================================================================
# function     GetUsedLanguages
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $aLanguageCodes = GetUsedLanguages( $aStoreNames );
# example      my $aLanguageCodes = GetUsedLanguages( [ keys %{ GetDBConnections( $DatabaseIni ) } ] );
#----------------------------------------------------------------------------------------
# description  returns the list of required language packs
#----------------------------------------------------------------------------------------
# input        $aStoreNames | list of databases | ref.array.string
# retun        $aLanguageCodes | list of iso639-2 language codes | ref.array.string
#========================================================================================
sub GetUsedLanguages {
    my ($aStoreNames) = @_;

    my %Languages;
    foreach my $StoreName (@$aStoreNames) {
        RunOnStore( 'Store'=> $StoreName, 'Sub' => sub {
                my $aaRows = GetCurrentDBHandle()->execute('SELECT code2 FROM language WHERE isused = 1');
                $Languages{$_} = 1 foreach map { $_->[0] } @$aaRows;
        });
    }

    return [ sort keys %Languages ];
}

############################################################################
# unpack language pack
############################################################################

sub TestLanguagePack {
    my ($HostMk, $PatchVersion) = @_;
    my $Errors;

    return $Errors if defined $HostMk->{LANGUAGES_OK};

    my ($Main,$Sub,$Level) = split(/\./, $PatchVersion);
    my $Version = "$Main${Sub}p$Level";

    my @Languages = grep { $_ !~ /^(de|en)$/ } split(/ /, $HostMk->{LANGUAGES});

    foreach my $Lang (@Languages) {
        print "TestLanguagePack: $Lang\n";
        my $LangDir = "langpack-$Lang-epages$Version";
        my $LangPack = -f "$LangDir.tar.gz" ? "$LangDir.tar.gz" : "../$LangDir.tar.gz";
        my $cmd;

        if (-f $LangPack) {
            # tar is much faster and requires much less memory than Archive::Tar
            if ($Config{osname} eq 'MSWin32') {
                $LangPack =~ s,/,\\,;
                $cmd = "gzip -cd \"$LangPack\" > \"$LangDir.tar\"";
                system($cmd);
                $Errors .= "$cmd: $?\n" if $?;
                $cmd = "tar xf \"$LangDir.tar\"";
                system($cmd);
                $Errors .= "$cmd: $?\n" if $?;
                unlink("$LangDir.tar");
            } else {
                $cmd = "tar zxf \"$LangPack\"";
                system($cmd);
                $Errors .= "$cmd: $?\n" if $?;
            }
        }

        if (-d "$LangDir/Cartridges") {
            CopyDirectory("$LangDir/Cartridges", "Cartridges");
        } else {
            $Errors .= "\nERROR: Required language pack failed: $LangDir.tar.gz\n";
        }
    }

    $HostMk->{LANGUAGES_OK} = $Errors ? 0 : 1;
    return $Errors;
}

############################################################################
# write makefile
############################################################################

sub WriteHostMk {
    my ($HostMk, $FirstArg) = @_;

    if ($Config{osname} eq 'MSWin32' and not defined $HostMk->{WINDOWS2008}) {
        my $VersionNumber = `ver`;
        chomp($VersionNumber);
        $VersionNumber =~ s/.*\[.*?([0-9][0-9.]*).*/$1/s;
        $HostMk->{WINDOWS2008} = $VersionNumber =~ /^6\./ ? 1 : "";
    } elsif ($Config{osname} ne 'MSWin32' and not defined $HostMk->{SU_EXE}) {
        $HostMk->{SU_EXE} = system("runuser -c true") ? 'su' : 'runuser';
    }

    $HostMk->{IS_ALPHA} = 1 if $FirstArg =~ /^(install|restore)$/;
    $HostMk->{CTMPL_TXT} = 1 if scalar glob("ctmpl-*.txt");

    my $HostName = hostname();
    my $t = time;
    open my $HOST_MK,">", "host-$HostName.mk" or die "host-$HostName.mk: $!";
    map { print $HOST_MK "$_=$HostMk->{$_}\n"; } sort keys %$HostMk;
    close($HOST_MK);
    utime($t,$t,"host-$HostName.mk");
}

############################################################################
# run make
############################################################################

sub GetMake {
    my ($SetEnv) = @_;

    if ($Config{osname} eq 'MSWin32') {
        my $Command = "nmake /? >nul 2>&1";
        print "$Command\n";
        system($Command);
        return 'nmake /nologo' unless $?;
        $Command = "$ENV{MAKE} /? >nul 2>&1";
        print "$Command\n";
        system($Command);
        return $ENV{MAKE} unless $?;
    } else {
        foreach my $Make (("make", $ENV{MAKE})) {
            my $Command = "$SetEnv $Make -v 2>&1 | grep GNU >/dev/null 2>&1 && echo $Make";
            print "$Command\n";
            my $MakeExe = `$Command`;
            chomp($MakeExe);
            return $MakeExe if $MakeExe;
        }
    }
    my $MakeType = $Config{osname} eq 'MSWin32' ? 'nmake' : 'GNU make';
    die "\nERROR: $MakeType: neither found in path, nor in envvar MAKE\n";
}

sub GetVar {
    my ($SetEnv, $MakeExe) = @_;

    my $HostName = hostname();
    my $Command = "$$SetEnv $MakeExe get_var HOSTNAME=$HostName";
    print "$Command\n";
    my $Result = `$Command`;
    print $Result;
    die if $?;

    my ($LogDir, $BakDir, $TgtDir, $Tgt);
    foreach my $Line (split(/\n/, $Result)) {
        if ($Line =~ /LOG_DIR=(.+)/) { $LogDir = $1; }
        if ($Line =~ /BACKUP_DIR=(.+)/) { $BakDir = $1; }
        if ($Line =~ /TGT_DIR=(.+)/) { $TgtDir = $1; }
        if ($Line =~ /TGT=(.+)/) { $Tgt = $1; }
    }

    if ($Config{'osname'} eq 'MSWin32') {
        $$SetEnv = "set EPAGES_BACKUP_DIR=$BakDir& ";
    } else {
        $$SetEnv .= "export EPAGES_BACKUP_DIR=$BakDir ; ";
    }

    foreach my $Dir ($LogDir, $TgtDir, $Tgt) {
        next if $Dir !~ /\S/ or -d $Dir;
        mkpath($Dir, 0, oct(755)) or die "\nERROR: $Dir: Can't create directory: $!";
    }
    return ($LogDir, $TgtDir);
}

sub DateString {
    my @Month = qw( Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec );
    my %date = ();
    my ($mon, $year);
    ($date{sec}, $date{min}, $date{hour}, $date{mday}, $mon, $year) = localtime(time);
    map { $date{$_} = "0$date{$_}" if $date{$_} < 10; } qw(sec min hour mday);
    return ($year + 1900). "-$Month[$mon]-$date{mday}-$date{hour}.$date{min}.$date{sec}"
}

sub GetHostArguments {
    my ($TgtDir) = @_;

    my ($PostInstall, $AlphaHost) = ("", "");

    if ( $TgtDir and -s "$TgtDir/alphahost.txt") {
        open my $FOO, "<", "$TgtDir/alphahost.txt" or die "$TgtDir/alphahost.txt: Can't read file: $!";
        $AlphaHost = <$FOO>;
        close $FOO;
        chomp($AlphaHost);
    }

    if ( -f "$TgtDir/prehosts.txt" ) {
        my @PostHosts = ();
        if ( -f "$TgtDir/posthosts.txt" ) {
            open my $FOO, "<", "$TgtDir/posthosts.txt" or die "$TgtDir/posthosts.txt: Can't read file: $!";
            @PostHosts = <$FOO>;
            chomp($_) foreach @PostHosts;
            close $FOO;
        }

        $PostInstall = "POSTINSTALL_COMPLETE=1";
        open my $FOO, "<", "$TgtDir/prehosts.txt" or die "$TgtDir/prehosts.txt: Can't read file: $!";
        foreach my $Host (<$FOO>) {
            chomp($Host);
            next if $Host eq $AlphaHost;
            next if $Host ~~ @PostHosts;
            $PostInstall = "";
            last;
        }
        close $FOO;
    }
    return $PostInstall;
}

sub RunMake {
    my (@ArgList) = @_;

    my ($FirstArg, $Arguments, $SetEnv) = GetArguments(@ArgList);
    my $MakeExe = GetMake($SetEnv);
    my ($LogDir, $TgtDir) = GetVar(\$SetEnv, $MakeExe);
    my $HostName = hostname();

    my $LogFile = "$LogDir/". DateString(). "-$FirstArg.log";
    my $PerlArgs = "-IInclude Scripts/SaveOutput.pl -log \"$LogFile\" -pid $$";
    my $PerlExe = $Config{'osname'} eq 'MSWin32' ? $^X : $ENV{'PERL'};
    my $HostArgs = GetHostArguments($TgtDir);
    my $Command = "$SetEnv ($MakeExe $Arguments $HostArgs HOSTNAME=$HostName 2>&1 || echo TARGET FAILED [PID: $$]) | $PerlExe $PerlArgs";

    open my $FOO, ">", $LogFile or die "\nERROR: $LogFile: Can't write file: $!";
    print $FOO "\n==========================================================\n";
    print $FOO "$Command\n";
    print $FOO "==========================================================\n\n";
    close $FOO;

    print "$Command\n";

    my $Time = time;
    system($Command);
    my $Seconds = time - $Time;

    my ($Minutes, $Hours) = (0, 0);

    if ($Seconds > 60) {
        $Minutes = int $Seconds / 60;
        $Seconds = $Seconds % 60;
    }
    if ($Minutes > 60) {
        $Hours = int $Minutes / 60;
        $Minutes = $Minutes % 60;
    }
    my $Duration = "";
    $Duration .= " $Hours hours," if $Hours;
    $Duration .= " $Minutes minutes," if $Minutes;
    $Duration .= " $Seconds seconds";

    open $FOO, ">>", $LogFile or die "\nERROR: $LogFile: Can't write file: $!";
    print $FOO "\n==========================================================\n";
    print $FOO "need$Duration\n";
    print $FOO "==========================================================\n\n";
    close $FOO;

    print "need$Duration\n";

    die if $?;
}

RunScript(
    'Sub' => \&Main
);
